(* ar.ml - encrypted archive files
   (C) 2007 Antoine Mine'
 *)


let debug = false


(* ARCHIVE *)
(* ******* *)

(* recursively create a directory *)
let rec mkdir_p pathname mode =
  if pathname="." || pathname="" then () 
  else (
    mkdir_p (Filename.dirname pathname) mode;
    try Unix.mkdir pathname mode with Unix.Unix_error (Unix.EEXIST,_,_) -> ()
   )

    
(* remove trailing blanks *)
let unblank s =
  let x = ref (String.length s) in
  while !x>0 && s.[!x-1]=' ' do decr x done;
  String.sub s 0 !x
    

(* make a (recursive) archive of files, and outputs it through out_func *)
let tar (filenames:string list) (out_func:string->unit) =
  let don = Hashtbl.create 2 in
  let rec add pathname =
    (* avoid infinite loops *)
    if Hashtbl.mem don pathname then () else
    try
      Hashtbl.add don pathname "blurg!";
      let s = Unix.lstat pathname in
      match s.Unix.st_kind with
      | Unix.S_LNK -> 
	  let link = Unix.readlink pathname in
	  out_func (Printf.sprintf "L%-512s%-512s" pathname link);
	  if debug then Printf.printf "adding link %s -> %s\n" pathname link;
	  (* recurse *)
	  add (Filename.concat (Filename.dirname pathname) link)
      | Unix.S_DIR -> 
	  out_func (Printf.sprintf "D%-512s%06i" pathname s.Unix.st_perm);
	  if debug then Printf.printf "adding dir %s %03o\n" pathname s.Unix.st_perm;
	  (* recurse *)
	  let a = Sys.readdir pathname in
	  Array.sort compare a;
	  Array.iter 
	    (fun f -> add (Filename.concat pathname f)) a
      | Unix.S_REG -> 
	  let f = open_in_bin pathname 
	  and b = String.create 65536 in
	  out_func 
	    (Printf.sprintf "F%-512s%06i%010i" pathname s.Unix.st_perm s.Unix.st_size);
	  if debug then Printf.printf "adding file %s %03o %i\n" pathname s.Unix.st_perm s.Unix.st_size;
	  (try while true do match input f b 0 65536 with
	  | 0 -> raise End_of_file
	  | x when x=String.length b -> out_func b
	  | x -> out_func (String.sub b 0 x)
	  done with End_of_file -> ());
	  close_in f
      | _ -> Printf.eprintf "skipping %s\n" pathname
    with Unix.Unix_error _ | Sys_error _ -> Printf.eprintf "skipping %s\n" pathname
  in
  List.iter add filenames;
  out_func "E"


(* read from in_fun and extract archive *)
let untar (in_func:int->string) =
  try while true do
    match in_func 1 with
    | "D" -> 
	let n = unblank (in_func 512) in
	let m = int_of_string (in_func 6) in
	if debug then Printf.printf "extracting directory %s %03o\n" n m;
	mkdir_p n m
    | "L" ->
	let n = unblank (in_func 512) in
	let l = unblank (in_func 512) in
	if debug then Printf.printf "extracting link %s -> %s\n" n l;
	mkdir_p (Filename.dirname n) 0o644;
	(try Unix.symlink l n 
	with Unix.Unix_error _ -> Printf.eprintf "skipping %s\n" n)
    | "F" ->
	let n = unblank (in_func 512) in
	let m = int_of_string (in_func 6) in
	let s = int_of_string (in_func 10) in
	if debug then Printf.printf "extracting file %s %03o %i\n" n m s;
	mkdir_p (Filename.dirname n) m;
	let f = open_out_bin n in
	for i=1 to s/65536 do output_string f (in_func 65536) done;
	if s mod 65536 <> 0 then output_string f (in_func (s mod 65536));
	close_out f;
	Unix.chmod n m
    | "E" -> raise End_of_file
    | _ -> failwith "Invalid archive"
  done with End_of_file -> ()


(* ENCRYPTION *)
(* ********** *)

(* random *)
let rand = [|
  0x89; 0x7D; 0x66; 0xEF; 0xEE; 0x5F; 0xB8; 0xCE; 
  0x04; 0x00; 0xAF; 0x14; 0x88; 0x36; 0x3A; 0xD3; 
  0xCB; 0x9A; 0x5A; 0x2A; 0x0B; 0x78; 0xB1; 0x6A; 
  0xEA; 0xCF; 0xD7; 0xD6; 0xA0; 0xF4; 0xB3; 0x08; 
  0x3B; 0x2B; 0x23; 0xB8; 0x3E; 0x0E; 0xA8; 0x3E; 
  0xA9; 0x63; 0x3A; 0x1E; 0x77; 0xD4; 0xC7; 0xFB; 
  0x8F; 0xFB; 0xD6; 0x46; 0x23; 0xFF; 0x8F; 0xDC; 
  0x44; 0x75; 0x0E; 0x22; 0x8F; 0x1D; 0x81; 0x52; 
  0xA9; 0xF2; 0x1E; 0x9D; 0xE7; 0xCC; 0x3C; 0xAF; 
  0x22; 0x1B; 0x25; 0x5F; 0x3C; 0x3A; 0x57; 0x4D; 
  0xBD; 0x4E; 0x15; 0xB2; 0x86; 0xE7; 0xF6; 0xA7; 
  0xFF; 0x69; 0x53; 0xAD; 0x87; 0x27; 0x87; 0xD6; 
  0x25; 0x9D; 0x2E; 0x92; 0x6F; 0xBA; 0x7C; 0x7E; 
  0x06; 0x97; 0x40; 0x40; 0xA0; 0x8E; 0x17; 0xE7; 
  0x03; 0xFE; 0x8B; 0x5C; 0x9A; 0x4D; 0x13; 0x60; 
  0xD1; 0x9D; 0x6C; 0x7A; 0x47; 0x8D; 0xF2; 0xFE; 

  0xC6; 0x12; 0x03; 0x7D; 0x49; 0xDF; 0x83; 0xD7; 
  0xBB; 0xC5; 0x09; 0xAA; 0xD0; 0x55; 0x34; 0xB5; 
  0x46; 0x43; 0x37; 0x58; 0x31; 0xCA; 0xFB; 0x0B; 
  0x44; 0x0D; 0x67; 0xFF; 0xF7; 0x9B; 0x14; 0x10; 
  0xB8; 0xC1; 0x36; 0xF2; 0x99; 0x9E; 0x6B; 0xEF; 
  0xF4; 0x14; 0xA5; 0xB7; 0xAC; 0x56; 0x86; 0x55; 
  0x47; 0x2A; 0x72; 0xEB; 0x0D; 0x53; 0x49; 0x46; 
  0xEB; 0x4D; 0x9A; 0x26; 0xD6; 0xAA; 0xD8; 0xE0; 
  0xDB; 0x28; 0x25; 0xB8; 0x43; 0x7F; 0xA8; 0x74; 
  0x9F; 0xB7; 0x43; 0x1D; 0x38; 0xD7; 0x08; 0x50; 
  0xC3; 0x06; 0x7B; 0xEB; 0x94; 0x01; 0xDC; 0xC8; 
  0xFB; 0x91; 0xC3; 0xA4; 0xB0; 0x47; 0x40; 0x8C; 
  0xC6; 0xEE; 0xCE; 0x52; 0xD2; 0x63; 0x9D; 0x6D; 
  0x8B; 0xEA; 0x78; 0x9F; 0x7A; 0xE8; 0x20; 0x8A; 
  0x40; 0xEC; 0x33; 0xEE; 0x71; 0x91; 0x48; 0x5B; 
  0x57; 0x27; 0xE8; 0x65; 0x20; 0x17; 0x90; 0xD0; 
|]

(* multiplication in Z[X] / X^8+X^4+X^3+X+1 *)
let mul_gf8 a b =
  assert(a<256);
  let rec toto b =
    if b=0 then 0 else
    let r = ((toto (b lsr 1)) * 2) lxor (if b land 1 = 0 then 0 else a) in
    if r<256 then r else r lxor 0x11b 
  in toto b

(* tabulate *)
let mul_gf8_table =
  Array.init 256 (fun x -> Array.init 256 (fun y -> mul_gf8 x y))
   

(* create a crypted file; returns a string write and a close function *)
let open_out_crypt fn : (string->unit)*(unit->unit) =
  let f = open_out_bin fn
  and pos = ref 0 
  and a = Array.copy rand in
  (fun s -> 
    for i=0 to String.length s - 1 do
      s.[i] <- Char.chr ((Char.code s.[i]) lxor (a.(!pos)));
      a.(!pos) <- mul_gf8_table.(Char.code s.[i]).(a.(!pos));
      pos := (!pos + 1) land 255;
    done;
    output_string f s),
  (fun () -> close_out f)

(* read a crypted file; returns a string read function and a close function *)
let open_in_crypt fn : (int->string)*(unit->unit) =
  let f = open_in_bin fn
  and pos = ref 0
  and a = Array.copy rand in
  (fun nb -> 
    let s = String.create nb in
    really_input f s 0 nb;
     for i=0 to String.length s - 1 do
      let olds = s.[i] in
      s.[i] <- Char.chr ((Char.code s.[i]) lxor (a.(!pos)));
      a.(!pos) <- mul_gf8_table.(Char.code olds).(a.(!pos));
      pos := (!pos + 1) land 255;
    done;
    s),
  (fun () -> close_in f)


(* MAIN FUNCTIONS *)
(* ************** *)

let archive (archivename:string) (filenames:string list) =
  if debug then Printf.printf "creating archive %s\n" archivename;
  let wr,cl = open_out_crypt archivename in
  tar filenames wr;
  cl();
  if debug then Printf.printf "archive created\n"

let unarchive (archivename:string) =
  if debug then Printf.printf "opening archive %s\n" archivename;
  let rd,cl = open_in_crypt archivename in
  untar rd;
  cl();
  if debug then Printf.printf "archive extracted\n"
    
    
	
