(***********************************************************************)
(*                                                                     *)
(*                           Objective Caml                            *)
(*                                                                     *)
(*            Xavier Leroy, projet Cristal, INRIA Rocquencourt         *)
(*                                                                     *)
(*  Copyright 1996 Institut National de Recherche en Informatique et   *)
(*  en Automatique.  All rights reserved.  This file is distributed    *)
(*  under the terms of the GNU Library General Public License, with    *)
(*  the special exception on linking described in file ../LICENSE.     *)
(*                                                                     *)
(***********************************************************************)

(** Cache based on hash-tables *)
(** Derived from the Hashtbl implementation in the OCaml 3.06 distribution *)
(** Modifications, Copyright (C) Antoine Mine' 2003 *)


(** Weak arrays that can exceed Sys.max_array_length = 4194303 *)
(** Implemented using an array of small weak arrays! *)
module Meta = 
  struct

    type 'a t = 'a Weak.t array

    (** size of inner arrays *)
    let max_size = 1024 * 1024	

    let create sz =
      Array.init ((sz+max_size-1)/max_size) 
	(fun i -> Weak.create (min max_size (sz-i*max_size)))

    let get a i   = Weak.get a.(i/max_size) (i mod max_size)
    let set a i d = Weak.set a.(i/max_size) (i mod max_size) d

    let length a = let l = Array.length a - 1 in l*max_size + Weak.length a.(l)

  end




type ('a, 'b) t =
  { mutable size: int;                        (* number of elements *)
    mutable data: ('a, 'b) bucketlist Meta.t } (* the buckets *)

and ('a, 'b) bucketlist = Cons of 'a * 'b * ('a, 'b) bucketlist option



module type S =
  sig
    type key
    type 'a t
    val create: int -> 'a t
    val clear: 'a t -> unit
    val add: 'a t -> key -> 'a -> unit
    val remove: 'a t -> key -> unit
    val find: 'a t -> key -> 'a
    val mem: 'a t -> key -> bool
  end

module Make(H: Hashtbl.HashedType): (S with type key = H.t) =
  struct
    type key = H.t
    type 'a hashtbl = (key, 'a) t
    type 'a t = 'a hashtbl

    let create initial_size =
      let s = max 1 initial_size in
      { size = 0; data = Meta.create s }
	
    let clear h =
      for i = 0 to Meta.length h.data - 1 do
	Meta.set h.data i None
      done;
      h.size <- 0


    let safehash key = (H.hash key) land max_int

    let resize hashfun tbl =
      let newsize = ref 0 in
      let odata = tbl.data in
      let osize = Meta.length odata in
      let nsize = 2 * osize + 1 in
      if nsize <> osize then begin
	let ndata = Meta.create nsize in
	let rec insert_bucket = function
            None -> ()
	  | Some (Cons(key, data, rest)) ->
	      incr newsize;
              insert_bucket rest; (* preserve original order of elements *)
              let nidx = (hashfun key) mod nsize in
              Meta.set ndata nidx (Some (Cons(key, data, Meta.get ndata nidx)))
	in
	for i = 0 to osize - 1 do
	  insert_bucket (Meta.get odata i)
	done;
	tbl.data <- ndata;
	tbl.size <- !newsize
      end

    let add h key info =
      let i = (safehash key) mod (Meta.length h.data) in
      let bucket = Cons(key, info, Meta.get h.data i) in
      Meta.set h.data i (Some bucket);
      h.size <- succ h.size;
      if h.size > Meta.length h.data lsl 1 then resize safehash h

    let remove h key =
      let rec remove_bucket = function
          None -> None
        | Some (Cons(k, i, next)) ->
            if H.equal k key
            then begin h.size <- pred h.size; next end
            else Some (Cons(k, i, remove_bucket next)) in
      let i = (safehash key) mod (Meta.length h.data) in
      Meta.set h.data i (remove_bucket (Meta.get h.data i))

    let rec find_rec key = function
        None -> raise Not_found
      | Some (Cons(k, d, rest)) ->
          if H.equal key k then d else find_rec key rest

    let find h key =
      match Meta.get h.data ((safehash key) mod (Meta.length h.data)) with
        None -> raise Not_found
      | Some (Cons(k1, d1, rest1)) ->
          if H.equal key k1 then d1 else
          match rest1 with
            None -> raise Not_found
          | Some (Cons(k2, d2, rest2)) ->
              if H.equal key k2 then d2 else
              match rest2 with
                None -> raise Not_found
              | Some (Cons(k3, d3, rest3)) ->
                  if H.equal key k3 then d3 else find_rec key rest3

    let mem h key =
      let rec mem_in_bucket = function
      | None -> false
      | Some (Cons(k, d, rest)) ->
          H.equal k key || mem_in_bucket rest in
      mem_in_bucket (Meta.get h.data ((safehash key) mod (Meta.length h.data)))


  end
