type computation_host = {
    host_address : Unix.sockaddr option;
    host_speed : float };;

let remove_comments =
  let comment_regexp = Str.regexp "[ \t]*\\(#.*\\)?$" in
  Str.replace_first comment_regexp "";;

let split_at_spaces =
  let split_regexp = Str.regexp "[ \t]+" in
  Str.split split_regexp;;

exception Syntax_error of string * int;;

let load_computation_hosts_from_file default_port filename =
  let in_chan = open_in filename
  and line_nr = ref 0
  and host_list = ref [] in
  let add_to_host_list address speed =
  host_list :=
    { host_address =
        if address="<self>"
        then None
        else Some (RPC.get_inet_address_and_port default_port address);
      host_speed = speed } :: !host_list in
  
  (try
    while true
    do
      let line = input_line in_chan in
      incr line_nr;
      match split_at_spaces (remove_comments line) with
      | [name; speed] ->
	  add_to_host_list name (float_of_string speed)
      | [] -> ()
      | _ ->  raise (Syntax_error(filename,!line_nr))
    done
  with End_of_file -> ());
  close_in in_chan;
  !host_list;;

let allocate_arrays hosts tasks =
  let n = Array.length hosts in
  assert (n > 0);
  let share_loads = Array.create n 0.
  and shares = Array.create n 0 in
  for i=1 to tasks
  do
    let best_time = ref infinity
    and chosen = ref (-1)in
    for j=0 to (pred n)
    do
      let cur_time = share_loads.(j) +. 1. /. hosts.(j).host_speed in
      if cur_time < !best_time
      then
        begin
	  best_time := cur_time;
	  chosen := j
        end
    done;
    shares.(!chosen) <- succ shares.(!chosen);
    share_loads.(!chosen) <- !best_time
  done;
  shares;;
  
let allocate_lists hosts tasks =
  Array.to_list (allocate_arrays (Array.of_list hosts) tasks)

  
let divide_workload : computation_host list -> 'a list -> 'a list list =
  fun hosts workload ->
  (* let workload = MoreList.shuffle_list workload in *)
  MoreList.split_list_by_numbers (allocate_lists hosts (List.length workload)) workload;;


let extract_local_and_remote x = 
  let rec toto accuml accumr = function
    | (Some addr, workload)::h -> toto accuml ((addr,workload)::accumr) h
    | (None, workload)::h -> toto (workload@accuml) accumr h
    | [] -> accuml,accumr
  in
  toto [] [] x

let parallel_call first_param f token hosts worklist =
  let work_shares = divide_workload hosts worklist in
  let local, remote = extract_local_and_remote
      (List.combine
	 (List.map (fun host -> host.host_address) hosts) work_shares) in
  let remote = 
    List.fold_right 
      (fun ((_,work) as pair) accum ->
	if work=[] then accum else pair::accum) remote []
  in
  let local_work, remote_work =
    RPC.rpc_parallel_call token
      (List.map (fun (addr,workload) -> addr,(first_param,workload)) remote)
      (fun () -> if local=[] then None else Some (f (first_param,local)) )
  in
  match local_work with Some x -> x::remote_work | None -> remote_work

let ping token hosts =
  let l = List.fold_left 
      (fun accum a -> 
	match a.host_address with Some x -> x::accum | None -> accum)
      [] hosts
  in
  RPC.rpc_ping token l

