/* IEEE floating-point */

/* C99 for exp2/log2; needs to be set *before* including math.h */
#define _ISOC99_SOURCE
#include <math.h>

/* We know GLIBC only after including math.h */
#if defined(__GLIBC__) && defined(__amd64)\
    && (__GLIBC__==2) && (__GLIBC_MINOR__ <= 3)
#define HAS_BUGGY_POW
#endif

#include <caml/mlvalues.h>
#include <caml/fail.h>
#include <caml/memory.h>
#include <caml/alloc.h>

/* Check for C99 */
#ifdef __STDC__
#if __STDC_VERSION__ >= 199901L
#define HAS_C99
#endif
#endif

/* Check for gcc-compatible extended inline asm */
#if defined(__INTEL_COMPILER)
#if __INTEL_COMPILER >= 500
#define HAS_EXTENDED_ASM
#endif
#endif

#ifdef __GNUC__
#define HAS_EXTENDED_ASM
#endif

#define SET_PREC_UNAVAILABLE

/* ISO C99 7.6.3.1 mandates fe{get|set}round in <fenv.h>.
   Linux (at least glibc) also supports it. */
#if defined(HAS_C99) || defined(__linux) || defined (__APPLE__)
#define FENV
#endif

/* FreeBSD 4.4 defines fp{get|set}round in <ieeefp.h>
Sun/Sparc used to define the same but no longer does so in SunOS 5.9 */
#if defined(__FreeBSD__) || defined(__sun)
#define IEEEFP
#endif

#ifdef FENV
#include <fenv.h>

#else /* defined(FENV) */

#ifdef IEEEFP
#include <ieeefp.h>

#if defined(__FreeBSD__)
#undef SET_PREC_UNAVAILABLE
#endif

#define fegetround fpgetround
#define fesetround fpsetround
#define FE_TONEAREST FP_RN
#define FE_UPWARD FP_RP
#define FE_DOWNWARD FP_RM
#define FE_TOWARDZERO FP_RZ

#define fegetprec fpgetprec
#define fesetprec fpsetprec
#define FE_FLTPREC FP_PS
#define FE_DBLPREC FP_PD
#define FE_LDBLPREC FP_PE

#else /* IEEEFP */
#define SET_ROUNDING_UNAVAILABLE
#define GET_ROUNDING_UNAVAILABLE
#endif /* IEEEFP */
#endif /* FENV */

#ifdef __ppc__
#define PPC_ASM
#undef SET_ROUNDING_UNAVAILABLE
#endif

#if (defined(__i386__) || defined(__amd64)) && defined(HAS_EXTENDED_ASM)
#define I386_ASM
#undef SET_PREC_UNAVAILABLE
#undef SET_ROUNDING_UNAVAILABLE
#if defined(__amd64) || !defined(PENTIUM_III_COMPATIBLE)
#define USE_SSE
#endif
#endif

value
mlfp_IEEE_setround(value rounding_mode)
{
  CAMLparam1(rounding_mode);

#ifdef SET_ROUNDING_UNAVAILABLE
  failwith("Fp_IEEE.setround unavailable");
#else

#ifdef I386_ASM
  static const int modes[] = { 0, 2, 1, 3 };

  {
    unsigned short cw;
    asm volatile ("fstcw %0" : "=m" (cw));
    cw = (cw & 0xF3FFU) | (modes[rounding_mode>>1] << 10);
    asm volatile ("fldcw %0" : : "m" (cw));
  }

#ifdef USE_SSE
  {
    unsigned long cw;
    asm volatile ("stmxcsr %0" : "=m" (cw));
    cw = (cw & 0x9FFFUL) | (modes[rounding_mode>>1] << 13);
    asm volatile ("ldmxcsr %0" : : "m" (cw));
  }
#endif


#elif defined(PPC_ASM)
  switch (rounding_mode)
  {
  case 1:
    	asm volatile ("mtfsfi 7,0");
  	break;
  case 3:
	asm volatile ("mtfsfi 7,2");
        break;
  case 5:
        asm volatile ("mtfsfi 7,3");
	break;
  case 7:
        asm volatile ("mtfsfi 7,1");
        break;
  default:
        failwith("Fp_IEEE.setround: unknown mode (typechecking error?)"); 
  }
#else
  /* if >= Pentium III, using SSE */

  static const int modes[] = {
    FE_TONEAREST, FE_UPWARD, FE_DOWNWARD, FE_TOWARDZERO };

  if (rounding_mode<1 || rounding_mode>7 || (rounding_mode&1) == 0)
    failwith("Fp_IEEE.setround: unknown mode (typechecking error?)"); 

  fesetround(modes[rounding_mode>>1]);
#endif
#endif
  CAMLreturn(Val_unit);
}

value
mlfp_IEEE_getround(value dummy)
{
  CAMLparam1(dummy);
#if defined(GET_ROUNDING_UNAVAILABLE)
  failwith("Fp_IEEE.getround unavailable");
#else
  value ret;
  switch (fegetround())
  {
  case FE_TONEAREST: ret=1;
  case FE_UPWARD: ret=3;
  case FE_DOWNWARD: ret=5;
  case FE_TOWARDZERO: ret=7;
  default:
    failwith("Fp_IEEE.getround: unknown mode (libc)"); 
  }
  CAMLreturn(ret);
#endif
}

value mlfp_IEEE_setprec(value precision)
{
  CAMLparam1(precision);
#ifdef SET_PREC_UNAVAILABLE
  failwith("Fp_IEEE.setprec unavailable");
#else

#ifdef I386_ASM
  static const unsigned short modes[] = { 0x000U, 0x200U, 0x300U };

  if (precision<1 || precision>5 || (precision&1)==0)
    failwith("Fp_IEEE.setprec: unknown mode (typechecking error?)");

  /* volatile is necessary for at least the Intel compiler v5.0. */
  {
    unsigned short cw;
    asm volatile ("fstcw %0" : "=m" (cw));
    cw = (cw & 0xFCFFU) | modes[precision>>1];
    asm volatile ("fldcw %0" : : "m" (cw));
  }
#else /* defined(I386_ASM) */

  static const int modes[] = {
    FE_FLTPREC, FE_DBLPREC, FE_LDBLPREC };

  if (precision<1 || precision>5 || (precision&1)==0)
    failwith("Fp_IEEE.setprec: unknown mode (typechecking error?)");

  fesetprec(modes[precision>>1]);
#endif /* defined(I386_ASM) */
#endif /* defined(PREC_UNAVAILABLE) */
  CAMLreturn(Val_unit);
}

#ifdef EXP2_LOG2
value mlfp_exp2(value x) {
  return caml_copy_double(exp2(Double_val(x)));
}                                                                                
value mlfp_log2(value x) {
  return caml_copy_double(log2(Double_val(x)));
}
#endif

value mlfp_pow(value x, value y) {
#ifdef HAS_BUGGY_POW
  return caml_copy_double(exp2(log2(Double_val(x))*Double_val(y)));
#else
  return caml_copy_double(pow(Double_val(x), Double_val(y)));
#endif
}                                                                

double mlfp_pow_f(double x, double y) {
#ifdef HAS_BUGGY_POW
  return exp2(log2(x)*y);
#else
  return pow(x,y);
#endif
}                                                                
