(**
Parallelization primitives
@author David Monniaux
*)

(*
Copyright (c) 2004 David Monniaux,
CNRS / Laboratoire d'informatique de l'Ecole normale superieure
*)


(** {2 Output utilities} *)

(** really flush *)
val flush: unit -> unit

(** diverts stdout and stderr to a temporaty file *)
val divert_stdout: unit -> (string*out_channel)
  
(** get the contents of stdout & stderr from the temporary file and destroy it *)
val readback_stdout: (string*out_channel) -> string


(**{2 Evaluation of pairs}

[xxxx_dual_eval f a g b] for any [xxx] evaluates identically as [(f a), (g b)] if [f] and [g] do not have side effects. *)

(** Normal sequential evaluation. *)
val normal_dual_eval : ('a -> 'b) -> 'a -> ('c -> 'd) -> 'c -> 'b * 'd

(*
(** Each element in the pair is evaluated in its own thread. *)
val threaded_dual_eval : ('a -> 'b) -> 'a -> ('c -> 'd) -> 'c -> 'b * 'd
*)

(** Whether to use pipes or temporary files for inter-process communications. Serialization to temporary files may use less temporary virtual memory. *)
type fork_mode =
    USE_PIPES
  | USE_TEMPORARY_FILES;;

(** Each element in the pair is evaluated in its own process. *)
val forked_dual_eval : fork_mode -> ('a -> 'b) -> 'a -> ('c -> 'd) -> 'c -> 'b * 'd

(**{2 List mappings}

[xxx_list_map f l] evaluates identically as [List.map f l] if [f] does not have side effects.*)

(*
(** Each element in the list is evaluated in its own thread. *)
val threaded_list_map : ('a -> 'b) -> 'a list -> 'b list
*)

(** Each element in the list is evaluated in its own process. *)
val forked_list_map : fork_mode -> ('a -> 'b) -> 'a list -> 'b list

(** [parallel_list_map n f l] splits the load on [n] processes. *)
val parallel_list_map : fork_mode -> int -> ('a list -> 'b) -> 'a list -> 'b list

(**{2 Associative operations on lists}

[xxxx_list_assoc_map neutral op f [l1; l2; l3...]] evaluates to
[(assoc (f l1) (assoc (f l2) (f .... (assoc (f ln) neutral))))]
if [f] has no side effects.
*)

(*
(** Each element in the list is evaluated in its own thread. *)
val threaded_list_assoc_map :
  'a -> ('b -> 'a -> 'a) -> ('c -> 'b) -> 'c list -> 'a
*)

(** Each element in the list is evaluated in its own process. *)
val forked_list_assoc_map : fork_mode -> 
  'a -> ('a -> 'b -> 'a) -> ('c -> 'b) -> 'c list -> 'a

(** [parallel_list_assoc_map neutral n op f l] divides the load on [n] processes. *)
val parallel_list_assoc_map : fork_mode -> int -> 'a -> ('a -> 'b -> 'a) -> ('c list -> 'b) -> 'c list -> 'a


(** {2 Fork memory optimization} *)

(**
   Issue: when forking two process, the memory consumption is doubled.
   Indeed, even memory shared by the processes is eventually fully copied as
   the GC disrupts the copy-on-write optimization.
   
   In fact, a large part of the memory copied in the forked process is useless.
   Indeed, as the forked process performs some computation and exists, a large
   part of its stack is inherited by the forking process and never accessed
   nor popped.

   Optimization: instead of executing the forked process's calculation on top
   of the forking process's stack, we launch a Execute exception that pop
   the whole stack, is caught at some top-level to perform the computation.
*)

(** Exception denoting some function to call before exiting. *)
exception Execute of (unit -> unit)

(** Wraps the function in the correct try ... with Execute ... statement *)
val catch_execute: ('a -> 'b) -> 'a -> 'b
