(*
 * Ptmap: Maps over integers implemented as Patricia trees.
 * Copyright (C) 2000 Jean-Christophe FILLIATRE
 * 
 * This software is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License version 2, as published by the Free Software Foundation.
 * 
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * 
 * See the GNU Library General Public License version 2 for more details
 * (enclosed in the file LGPL).
 *)

(*i $Id: ptmap.ml,v 1.1 2008/04/30 15:25:18 monniaux Exp $ i*)

(*s Maps of integers implemented as Patricia trees, following Chris
    Okasaki and Andrew Gill's paper {\em Fast Mergeable Integer Maps}
    ({\tt\small http://www.cs.columbia.edu/\~{}cdo/papers.html\#ml98maps}).
    See the documentation of module [Ptset] which is also based on the
    same data-structure. *)

type key = int

type 'a t =
  | Empty
  | Leaf of int * 'a
  | Branch of int * int * 'a t * 'a t

let empty = Empty

let is_empty t = t = Empty

let zero_bit k m = (k land m) == 0

let lowest_bit x = x land (-x)
let branching_bit p0 p1 = lowest_bit (p0 lxor p1)
let mask p m = p land (m-1)
let match_prefix k p m = (mask k m) == p
let is_strict_subpath p1 m1 p2 m2 =
  m1 > m2 && (match_prefix p1 p2 m2)

let rec mem k = function
  | Empty -> false
  | Leaf (j,_) -> k == j
  | Branch (_, m, l, r) -> mem k (if zero_bit k m then l else r)

let rec find k = function
  | Empty -> raise Not_found
  | Leaf (j,x) -> if k == j then x else raise Not_found
  | Branch (_, m, l, r) -> find k (if zero_bit k m then l else r)


let join (p0,t0,p1,t1) =
  let m = branching_bit p0 p1 in
  if zero_bit p0 m then 
    Branch (mask p0 m, m, t0, t1)
  else 
    Branch (mask p0 m, m, t1, t0)

let add k x t =
  let rec ins = function
    | Empty -> Leaf (k,x)
    | Leaf (j,_) as t -> 
	if j == k then Leaf (k,x) else join (k, Leaf (k,x), j, t)
    | Branch (p,m,t0,t1) as t ->
	if match_prefix k p m then
	  if zero_bit k m then 
	    Branch (p, m, ins t0, t1)
	  else
	    Branch (p, m, t0, ins t1)
	else
	  join (k, Leaf (k,x), p, t)
  in
  ins t

let branch = function
  | (_,_,Empty,t) -> t
  | (_,_,t,Empty) -> t
  | (p,m,t0,t1)   -> Branch (p,m,t0,t1)

let remove k t =
  let rec rmv = function
    | Empty -> Empty
    | Leaf (j,_) as t -> if k == j then Empty else t
    | Branch (p,m,t0,t1) as t -> 
	if match_prefix k p m then
	  if zero_bit k m then
	    branch (p, m, rmv t0, t1)
	  else
	    branch (p, m, t0, rmv t1)
	else
	  t
  in
  rmv t

let rec iter f = function
  | Empty -> ()
  | Leaf (k,x) -> f k x
  | Branch (_,_,t0,t1) -> iter f t0; iter f t1

let rec map f = function
  | Empty -> Empty
  | Leaf (k,x) -> Leaf (k, f x)
  | Branch (p,m,t0,t1) -> Branch (p, m, map f t0, map f t1)
      
let rec mapi f = function
  | Empty -> Empty
  | Leaf (k,x) -> Leaf (k, f k x)
  | Branch (p,m,t0,t1) -> Branch (p, m, mapi f t0, mapi f t1)
      
let rec fold f s accu = match s with
  | Empty -> accu
  | Leaf (k,x) -> f k x accu
  | Branch (_,_,t0,t1) -> fold f t0 (fold f t1 accu)

(* we order constructors as Empty < Leaf < Branch *)
let compare cmp t1 t2 =
  let rec compare_aux t1 t2 = match t1,t2 with
    | Empty, Empty -> 0
    | Empty, _ -> -1
    | _, Empty -> 1
    | Leaf (k1,x1), Leaf (k2,x2) ->
	let c = compare k1 k2 in 
	if c <> 0 then c else cmp x1 x2
    | Leaf _, Branch _ -> -1
    | Branch _, Leaf _ -> 1
    | Branch (p1,m1,l1,r1), Branch (p2,m2,l2,r2) ->
	let c = compare p1 p2 in
	if c <> 0 then c else
	let c = compare m1 m2 in
	if c <> 0 then c else
        let c = compare_aux l1 l2 in
        if c <> 0 then c else
        compare_aux r1 r2
  in
  compare_aux t1 t2

let equal eq t1 t2 =
  let rec equal_aux t1 t2 = match t1, t2 with
    | Empty, Empty -> true
    | Leaf (k1,x1), Leaf (k2,x2) -> k1 = k2 && eq x1 x2
    | Branch (p1,m1,l1,r1), Branch (p2,m2,l2,r2) ->
	p1 = p2 && m1 = m2 && equal_aux l1 l2 && equal_aux r1 r2
    | _ -> false
  in
  equal_aux t1 t2

let rec map2iz f a b =
  if a==b then a else
  match a,b with
  | Empty, Empty -> Empty
  | (Leaf (k1,x1)) ,(Leaf (k2,x2)) ->
      ((if k1 <> k2 then raise (Invalid_argument "map2iz: k1<>k2"));
      Leaf (k1, (f k1 x1 x2)))
  | (Branch (p1,m1,t10,t11)), (Branch (p2,m2,t20,t21)) ->
      ((if p1 <> p2 then raise (Invalid_argument "map2iz: p1<>p2"));
      (if m1 <> m2 then raise (Invalid_argument "map2iz: m1<>m2"));
      Branch (p1, m1, (map2iz f t10 t20), (map2iz f t11 t21)))
  | _ -> raise (Invalid_argument "map2iz: nodes not correlated");;

let rec map2 f1 f2 f a b =
  match a,b with
  | Empty, Empty -> Empty
  | Empty, _ -> mapi f2 b
  | _, Empty -> mapi f1 a

  | (Leaf (k1, x1)), (Leaf (k2, x2)) ->
      if k1=k2
      then Leaf (k1, (f k1 x1 x2))
      else
	let m = branching_bit k1 k2
	and t1 = Leaf (k1, (f1 k1 x1))
	and t2 = Leaf (k2, (f2 k2 x2)) in
	if zero_bit k1 m then 
	  Branch ((mask k1 m), m, t1, t2)
	else 
	  Branch ((mask k1 m), m, t2, t1)

  | (Branch(p1, m1, t10, t11)), (Branch (p2, m2, t20, t21)) ->
    if p1=p2 && m1=m2
    then Branch (p1, m1, (map2 f1 f2 f t10 t20), (map2 f1 f2 f t11 t21))
    else
      (if is_strict_subpath p1 m1 p2 m2
      then
        (if zero_bit p1 m2
	then Branch(p2, m2, (map2 f1 f2 f a t20), (mapi f2 t21))
	else Branch(p2, m2, (mapi f2 t20), (map2 f1 f2 f a t21)))
      else if is_strict_subpath p2 m2 p1 m1
      then
        (if zero_bit p2 m1
        then Branch(p1, m1, (map2 f1 f2 f t10 b), (mapi f1 t11))
	else Branch(p1, m1, (mapi f1 t10), (map2 f1 f2 f t11 b)))
      else 
	let m = branching_bit p1 p2
	and t1 = mapi f1 a
	and t2 = mapi f2 b in
	if zero_bit p1 m then 
	  Branch ((mask p1 m), m, t1, t2)
	else 
	  Branch ((mask p1 m), m, t2, t1))

  | (Leaf(k1, x1)), (Branch (p2, m2, t20, t21)) ->
      if match_prefix k1 p2 m2
      then
	(if zero_bit k1 m2
	then Branch(p2, m2, (map2 f1 f2 f a t20), (mapi f2 t21))
	else Branch(p2, m2, (mapi f2 t20), (map2 f1 f2 f a t21)))
      else
	let m = branching_bit k1 p2
	and t1 = Leaf(k1, (f1 k1 x1)) 
	and t2 = mapi f2 b in
 	if zero_bit k1 m then 
	  Branch ((mask k1 m), m, t1, t2)
	else 
	  Branch ((mask k1 m), m, t2, t1)

  | (Branch(p1, m1, t10, t11)), (Leaf(k2, x2)) ->
      if match_prefix k2 p1 m1
      then
	(if zero_bit k2 m1
	then Branch(p1, m1, (map2 f1 f2 f t10 b), (mapi f1 t11))
	else Branch(p1, m1, (mapi f1 t10), (map2 f1 f2 f t11 b)))
      else
	let m = branching_bit k2 p1
	and t1 = mapi f1 a
	and t2 = Leaf(k2, (f2 k2 x2)) in
 	if zero_bit p1 m then 
	  Branch ((mask k2 m), m, t1, t2)
	else 
	  Branch ((mask k2 m), m, t2, t1)

let rec map2i f1 f2 f a b =
  if a==b then a else
  match a,b with
  | Empty, Empty -> Empty
  | Empty, _ -> mapi f2 b
  | _, Empty -> mapi f1 a

  | (Leaf (k1, x1)), (Leaf (k2, x2)) ->
      if k1=k2
      then Leaf (k1, (f k1 x1 x2))
      else
	let m = branching_bit k1 k2
	and t1 = Leaf (k1, (f1 k1 x1))
	and t2 = Leaf (k2, (f2 k2 x2)) in
	if zero_bit k1 m then 
	  Branch ((mask k1 m), m, t1, t2)
	else 
	  Branch ((mask k1 m), m, t2, t1)

  | (Branch(p1, m1, t10, t11)), (Branch (p2, m2, t20, t21)) ->
    if p1=p2 && m1=m2
    then Branch (p1, m1, (map2i f1 f2 f t10 t20), (map2i f1 f2 f t11 t21))
    else
      (if is_strict_subpath p1 m1 p2 m2
      then
        (if zero_bit p1 m2
	then Branch(p2, m2, (map2i f1 f2 f a t20), (mapi f2 t21))
	else Branch(p2, m2, (mapi f2 t20), (map2i f1 f2 f a t21)))
      else if is_strict_subpath p2 m2 p1 m1
      then
        (if zero_bit p2 m1
        then Branch(p1, m1, (map2i f1 f2 f t10 b), (mapi f1 t11))
	else Branch(p1, m1, (mapi f1 t10), (map2i f1 f2 f t11 b)))
      else 
	let m = branching_bit p1 p2
	and t1 = mapi f1 a
	and t2 = mapi f2 b in
	if zero_bit p1 m then 
	  Branch ((mask p1 m), m, t1, t2)
	else 
	  Branch ((mask p1 m), m, t2, t1))

  | (Leaf(k1, x1)), (Branch (p2, m2, t20, t21)) ->
      if match_prefix k1 p2 m2
      then
	(if zero_bit k1 m2
	then Branch(p2, m2, (map2i f1 f2 f a t20), (mapi f2 t21))
	else Branch(p2, m2, (mapi f2 t20), (map2i f1 f2 f a t21)))
      else
	let m = branching_bit k1 p2
	and t1 = Leaf(k1, (f1 k1 x1)) 
	and t2 = mapi f2 b in
 	if zero_bit k1 m then 
	  Branch ((mask k1 m), m, t1, t2)
	else 
	  Branch ((mask k1 m), m, t2, t1)

  | (Branch(p1, m1, t10, t11)), (Leaf(k2, x2)) ->
      if match_prefix k2 p1 m1
      then
	(if zero_bit k2 m1
	then Branch(p1, m1, (map2i f1 f2 f t10 b), (mapi f1 t11))
	else Branch(p1, m1, (mapi f1 t10), (map2i f1 f2 f t11 b)))
      else
	let m = branching_bit k2 p1
	and t1 = mapi f1 a
	and t2 = Leaf(k2, (f2 k2 x2)) in
 	if zero_bit p1 m then 
	  Branch ((mask k2 m), m, t1, t2)
	else 
	  Branch ((mask k2 m), m, t2, t1)

let create_leaf_opt k = function
    None -> Empty
  | Some x -> Leaf(k, x)

let branch_opt (p, m, t1, t2) =
  match t1, t2 with
    Empty, _ -> t2
  | _, Empty -> t1
  | _ -> Branch(p, m, t1, t2)

let rec mapio f = function
  | Empty -> Empty
  | Leaf (k,x) -> create_leaf_opt k (f k x)
  | Branch (p,m,t0,t1) -> branch_opt (p, m, mapio f t0, mapio f t1)
      
let rec map2io f1 f2 f a b =
  if a==b then a else
  match a,b with
  | Empty, Empty -> Empty
  | Empty, _ -> mapio f2 b
  | _, Empty -> mapio f1 a

  | (Leaf (k1, x1)), (Leaf (k2, x2)) ->
      if k1=k2
      then create_leaf_opt k1 (f k1 x1 x2)
      else
	let m = branching_bit k1 k2
	and t1 = create_leaf_opt k1 (f1 k1 x1)
	and t2 = create_leaf_opt k2 (f2 k2 x2) in
	if zero_bit k1 m then 
	  branch_opt ((mask k1 m), m, t1, t2)
	else 
	  branch_opt ((mask k1 m), m, t2, t1)

  | (Branch(p1, m1, t10, t11)), (Branch (p2, m2, t20, t21)) ->
    if p1=p2 && m1=m2
    then branch_opt (p1, m1, (map2io f1 f2 f t10 t20),
		     (map2io f1 f2 f t11 t21))
    else
      (if is_strict_subpath p1 m1 p2 m2
      then
        (if zero_bit p1 m2
	then branch_opt(p2, m2, (map2io f1 f2 f a t20), (mapio f2 t21))
	else branch_opt(p2, m2, (mapio f2 t20), (map2io f1 f2 f a t21)))
      else if is_strict_subpath p2 m2 p1 m1
      then
        (if zero_bit p2 m1
        then branch_opt(p1, m1, (map2io f1 f2 f t10 b), (mapio f1 t11))
	else branch_opt(p1, m1, (mapio f1 t10), (map2io f1 f2 f t11 b)))
      else 
	let m = branching_bit p1 p2
	and t1 = mapio f1 a
	and t2 = mapio f2 b in
	if zero_bit p1 m then 
	  branch_opt ((mask p1 m), m, t1, t2)
	else 
	  branch_opt ((mask p1 m), m, t2, t1))

  | (Leaf(k1, x1)), (Branch (p2, m2, t20, t21)) ->
      if match_prefix k1 p2 m2
      then
	(if zero_bit k1 m2
	then branch_opt(p2, m2, (map2io f1 f2 f a t20), (mapio f2 t21))
	else branch_opt(p2, m2, (mapio f2 t20), (map2io f1 f2 f a t21)))
      else
	let m = branching_bit k1 p2
	and t1 = create_leaf_opt k1 (f1 k1 x1)
	and t2 = mapio f2 b in
 	if zero_bit k1 m then 
	  branch_opt ((mask k1 m), m, t1, t2)
	else 
	  branch_opt ((mask k1 m), m, t2, t1)

  | (Branch(p1, m1, t10, t11)), (Leaf(k2, x2)) ->
      if match_prefix k2 p1 m1
      then
	(if zero_bit k2 m1
	then branch_opt(p1, m1, (map2io f1 f2 f t10 b), (mapio f1 t11))
	else branch_opt(p1, m1, (mapio f1 t10), (map2io f1 f2 f t11 b)))
      else
	let m = branching_bit k2 p1
	and t1 = mapio f1 a
	and t2 = create_leaf_opt k2 (f2 k2 x2) in
 	if zero_bit p1 m then 
	  branch_opt ((mask k2 m), m, t1, t2)
	else 
	  branch_opt ((mask k2 m), m, t2, t1)

let rec iter2iz f a b =
  if a==b then () else
  match a,b with
    Empty, Empty -> ()
  | (Leaf (k1,x1)) ,(Leaf (k2,x2)) ->
      ((if k1 <> k2 then raise (Invalid_argument "iter2iz: k1<>k2"));
      (f k1 x1 x2))
  | (Branch (p1,m1,t10,t11)), (Branch (p2,m2,t20,t21)) ->
      ((if p1 <> p2 then raise (Invalid_argument "iter2iz: p1<>p2"));
      (if m1 <> m2 then raise (Invalid_argument "iter2iz: m1<>m2"));
      (iter2iz f t10 t20);
      (iter2iz f t11 t21))
  | _ -> raise (Invalid_argument "iter2iz: nodes not correlated");;

let rec iter2 (f1 : key -> 'a -> unit) (f2 : key -> 'b -> unit)
    (f : key -> 'a -> 'b -> unit) (a : 'a t) (b : 'b t) =
  match a,b with
  | Empty, Empty -> ()
  | Empty, _ -> iter f2 b
  | _, Empty -> iter f1 a

  | (Leaf (k1, x1)), (Leaf (k2, x2)) ->
      if k1=k2
      then f k1 x1 x2
      else
	(f1 k1 x1;
	f2 k2 x2)

  | (Branch(p1, m1, t10, t11)), (Branch (p2, m2, t20, t21)) ->
    if p1=p2 && m1=m2
    then ((iter2 f1 f2 f t10 t20); (iter2 f1 f2 f t11 t21))
    else
      (if is_strict_subpath p1 m1 p2 m2
      then
        (if zero_bit p1 m2
	then ((iter2 f1 f2 f a t20); (iter f2 t21))
	else ((iter f2 t20); (iter2 f1 f2 f a t21)))
      else if is_strict_subpath p2 m2 p1 m1
      then
        (if zero_bit p2 m1
        then ((iter2 f1 f2 f t10 b); (iter f1 t11))
	else ((iter f1 t10); (iter2 f1 f2 f t11 b)))
      else 
	(iter f1 a;
	iter f2 b))

  | (Leaf(k1, x1)), (Branch (p2, m2, t20, t21)) ->
      if match_prefix k1 p2 m2
      then
	(if zero_bit k1 m2
	then ((iter2 f1 f2 f a t20); (iter f2 t21))
	else ((iter f2 t20); (iter2 f1 f2 f a t21)))
      else
	(f1 k1 x1);
	(iter f2 b)

  | (Branch(p1, m1, t10, t11)), (Leaf(k2, x2)) ->
      if match_prefix k2 p1 m1
      then
	(if zero_bit k2 m1
	then ((iter2 f1 f2 f t10 b); (iter f1 t11))
	else ((iter f1 t10); (iter2 f1 f2 f t11 b)))
      else
	(iter f1 a);
	(f2 k2 x2)

let rec iter2i f1 f2 f a b =
  if a!=b then
  match a,b with
  | Empty, Empty -> ()
  | Empty, _ -> iter f2 b
  | _, Empty -> iter f1 a

  | (Leaf (k1, x1)), (Leaf (k2, x2)) ->
      if k1=k2
      then f k1 x1 x2
      else
	(f1 k1 x1;
	f2 k2 x2)

  | (Branch(p1, m1, t10, t11)), (Branch (p2, m2, t20, t21)) ->
    if p1=p2 && m1=m2
    then ((iter2i f1 f2 f t10 t20); (iter2i f1 f2 f t11 t21))
    else
      (if is_strict_subpath p1 m1 p2 m2
      then
        (if zero_bit p1 m2
	then ((iter2i f1 f2 f a t20); (iter f2 t21))
	else ((iter f2 t20); (iter2i f1 f2 f a t21)))
      else if is_strict_subpath p2 m2 p1 m1
      then
        (if zero_bit p2 m1
        then ((iter2i f1 f2 f t10 b); (iter f1 t11))
	else ((iter f1 t10); (iter2i f1 f2 f t11 b)))
      else 
	(iter f1 a;
	iter f2 b))

  | (Leaf(k1, x1)), (Branch (p2, m2, t20, t21)) ->
      if match_prefix k1 p2 m2
      then
	(if zero_bit k1 m2
	then ((iter2i f1 f2 f a t20); (iter f2 t21))
	else ((iter f2 t20); (iter2i f1 f2 f a t21)))
      else
	(f1 k1 x1);
	(iter f2 b)

  | (Branch(p1, m1, t10, t11)), (Leaf(k2, x2)) ->
      if match_prefix k2 p1 m1
      then
	(if zero_bit k2 m1
	then ((iter2i f1 f2 f t10 b); (iter f1 t11))
	else ((iter f1 t10); (iter2i f1 f2 f t11 b)))
      else
	(iter f1 a);
	(f2 k2 x2)

let rec fold2 (f1 : key -> 'a -> 'z -> 'z)
    (f2 : key -> 'b -> 'z -> 'z)
    (f : key -> 'a -> 'b -> 'z -> 'z) (a : 'a t) (b : 'b t) (z0 : 'z) =
  match a,b with
  | Empty, Empty -> z0
  | Empty, _ -> fold f2 b z0
  | _, Empty -> fold f1 a z0

  | (Leaf (k1, x1)), (Leaf (k2, x2)) ->
      if k1=k2
      then f k1 x1 x2 z0
      else
	(f1 k1 x1
	  (f2 k2 x2 z0))

  | (Branch(p1, m1, t10, t11)), (Branch (p2, m2, t20, t21)) ->
    if p1=p2 && m1=m2
    then (fold2 f1 f2 f t10 t20 (fold2 f1 f2 f t11 t21 z0))
    else
      (if is_strict_subpath p1 m1 p2 m2
      then
        (if zero_bit p1 m2
	then (fold2 f1 f2 f a t20 (fold f2 t21 z0))
	else (fold f2 t20 (fold2 f1 f2 f a t21 z0)))
      else if is_strict_subpath p2 m2 p1 m1
      then
        (if zero_bit p2 m1
        then (fold2 f1 f2 f t10 b (fold f1 t11 z0))
	else (fold f1 t10 (fold2 f1 f2 f t11 b z0)))
      else 
	(fold f1 a (fold f2 b z0)))

  | (Leaf(k1, x1)), (Branch (p2, m2, t20, t21)) ->
      if match_prefix k1 p2 m2
      then
	(if zero_bit k1 m2
	then (fold2 f1 f2 f a t20 (fold f2 t21 z0))
	else (fold f2 t20 (fold2 f1 f2 f a t21 z0)))
      else
	(f1 k1 x1
	  (fold f2 b z0))

  | (Branch(p1, m1, t10, t11)), (Leaf(k2, x2)) ->
      if match_prefix k2 p1 m1
      then
	(if zero_bit k2 m1
	then (fold2 f1 f2 f t10 b (fold f1 t11 z0))
	else (fold f1 t10 (fold2 f1 f2 f t11 b z0)))
      else
	(fold f1 a
	  (f2 k2 x2 z0))

let rec fold2i 
    (f : key -> 'a -> 'b -> 'z -> 'z) (a : 'a t) (b : 'b t) (z0 : 'z) =
  match a,b with
  | (Empty, _) | (_, Empty) -> z0

  | (Leaf (k1, x1)), (Leaf (k2, x2)) ->
      if k1=k2
      then f k1 x1 x2 z0
      else z0

  | (Branch(p1, m1, t10, t11)), (Branch (p2, m2, t20, t21)) ->
    if p1=p2 && m1=m2
    then (fold2i f t10 t20 (fold2i f t11 t21 z0))
    else
      (if is_strict_subpath p1 m1 p2 m2
      then
        (if zero_bit p1 m2
	then (fold2i f a t20 z0)
	else (fold2i f a t21 z0))
      else if is_strict_subpath p2 m2 p1 m1
      then
        (if zero_bit p2 m1
        then (fold2i f t10 b z0)
	else (fold2i f t11 b z0))
      else z0)

  | (Leaf(k1, x1)), (Branch (p2, m2, t20, t21)) ->
      if match_prefix k1 p2 m2
      then
	(if zero_bit k1 m2
	then (fold2i f a t20 z0)
	else (fold2i f a t21 z0))
      else z0

  | (Branch(p1, m1, t10, t11)), (Leaf(k2, x2)) ->
      if match_prefix k2 p1 m1
      then
	(if zero_bit k2 m1
	then (fold2i f t10 b z0)
	else (fold2i f t11 b z0))
      else z0


exception Found_item
let forall2iz p m1 m2 =
  try
    iter2iz (fun k1 x1 x2 ->
      if x1!=x2 && not (p k1 x1 x2) then raise Found_item) m1 m2;
    true
  with Found_item -> false;;

  type 'a delta_record = 
      ('a t) *   (* keys in m2, possibly not in m1 *)
      (key list) (* keys in m1 and not in m2 *)

  let compute_delta (equal:'a->'a->bool) (m1:'a t) (m2:'a t) =
    let to_add = ref empty
    and to_rem = ref [] in
    iter2i 
      (fun k v1 -> to_rem := k::(!to_rem))
      (fun k v2 -> to_add := add k v2 !to_add)
      (fun k v1 v2 ->
	if not (equal v1 v2) then to_add := add k v2 !to_add)
      m1 m2;
    !to_add, !to_rem

  let apply_delta (m1:'a t) ((add,rem):'a delta_record) =
    let m2 = 
      map2i
	(fun k v -> v)
	(fun k v -> v)
	(fun k v1 v2 -> v2)
	m1 add
    in 
    List.fold_left (fun m k -> remove k m) m2 rem

let rec size = function
    Empty -> 0
  | Leaf _ -> 1
  | Branch(_, _, l, r) -> (size l) + (size r)


(* TODO: implement iter_slice, fold_slice, mapi_slice *)
