let find_function_in_file (file : Cil.file)
    (function_name : string) : Cil.fundec option =
  let found = ref None in
    (try
      List.iter
	(function
	   | Cil.GFun(fundec, _) when fundec.Cil.svar.Cil.vname =
	       function_name ->
	       found := Some fundec;
	       raise Exit
	   | _ -> ())
	file.Cil.globals
    with Exit -> ());
    !found;;

let collect_variables (variables : Cil.varinfo list) =
  List.map
    (fun var -> var.Cil.vname)
    (List.filter
       (fun (varinfo : Cil.varinfo) ->
	  Cil.isArithmeticType varinfo.Cil.vtype)
       variables);;

let vmin tag v = v ^ "$min_" ^ tag and vmax tag v  = v ^ "$max_" ^ tag;;

let gen_conditions tag (vars : (string*string) list) :
    string list *
    Formula.source_unquantified_formula *
    Formula.source_unquantified_formula =
  (List.concat (List.map (fun (v, _) -> [vmin tag v; vmax tag v]) vars)),
  (Formula.mk_v_and (List.concat (List.map
     (fun (v,v2)-> [`Atomic(`Comparison((`Variable v2), `LESS_EQ, (`Variable (vmax tag v))));
               `Atomic(`Comparison((`Variable v2), `GREATER_EQ, (`Variable (vmin tag v))))]) vars) )),
  (Formula.mk_v_and (List.map
     (fun (v,_)-> `Atomic(`Comparison((`Variable (vmin tag v)), `LESS_EQ, (`Variable (vmax tag v))))) vars));;

let gen_preconditions vars = gen_conditions "in" (List.map (fun x -> x,x) vars)

module StringMap = Map2.Make(String)

type semantics = {
  direct_flow : Formula.source_unquantified_formula;
  break_flow : Formula.source_unquantified_formula
}

let skip_semantics = {
  direct_flow = Formula.formula_true;
  break_flow = Formula.formula_false
}

let break_semantics = {
  direct_flow = Formula.formula_false;
  break_flow = Formula.formula_true
}

let normal_semantics f = {
  direct_flow = f;
  break_flow = Formula.formula_false
}

let compose_semantics t1 t2 = {
  direct_flow = Formula.mk_and t1.direct_flow t2.direct_flow;
  break_flow = Formula.mk_or t1.break_flow (Formula.mk_and t1.direct_flow t2.break_flow)
};;

let meet_semantics sem1 sem2 = {
  direct_flow = Formula.mk_or sem1.direct_flow sem2.direct_flow;
  break_flow = Formula.mk_or sem1.break_flow sem2.break_flow
}

let filter_semantics sem f = {
  direct_flow = Formula.mk_and sem.direct_flow f;
  break_flow = Formula.mk_and sem.break_flow f
}

type translator_state = {
    variable_index : string StringMap.t;
    variable_counter : int
}

let empty_translator_state = {
  variable_index = StringMap.empty;
  variable_counter = 0
};;

let get_variable_name state vname =
  try StringMap.find vname state.variable_index
  with Not_found -> vname;;

let get_new_variable_name vname state location =
  let new_name = Printf.sprintf "%s#%d" vname state.variable_counter in
  {state with
   variable_counter = 1 + state.variable_counter;
   variable_index = StringMap.add vname new_name state.variable_index },
  new_name;;

let get_tmp_variable_name vname state =
  {state with variable_counter = 1 + state.variable_counter},
  (Printf.sprintf "%s#%d" vname state.variable_counter);;

let get_top_boolean state =
  let state', v = get_tmp_variable_name "topBool" state in
    state',
  (`Atomic(`Comparison((`Variable v), `GREATER, (`Integer Z.zero)))),
  (`Atomic(`Comparison((`Variable v), `LESS_EQ, (`Integer Z.zero)))),
  Formula.formula_true

let merge_translator_states location s1 s2 =
  let f1 = ref[]
  and f2 = ref[]
  and cur_state = ref s2 in
  let new_index =
    StringMap.map2i (fun v i1 -> i1) (fun v i2 -> i2)
      (fun v i1 i2 ->
	let new_state, i' = get_new_variable_name v !cur_state location in
	cur_state := new_state;
        let v1 = `Variable i1
        and v2 = `Variable i2
        and v' = `Variable i' in
	f1 := (`Atomic(`Comparison(v', `EQUAL, v1))) :: !f1;
	f2 := (`Atomic(`Comparison(v', `EQUAL, v2))) :: !f2;
	i'
      )
      s1.variable_index s2.variable_index in
    { !cur_state with variable_index = new_index}, !f1, !f2;;

let swap (a, b) = (b, a)

let gen_integer_leq v1 v2 =
  (`Atomic(`Comparison(v1, `LESS_EQ, v2))),
	(`Atomic(`Comparison(v1, `GREATER_EQ,
			     (Formula.mk_sum v2 (`Integer Z.one)))))

let gen_integer_equal v1 v2 =
  (`Atomic(`Comparison(v1, `EQUAL, v2))),
  (Formula.mk_or
     (`Atomic (`Comparison(v1, `GREATER_EQ,
		  (Formula.mk_sum v2 (`Integer Z.one)))))
     (`Atomic (`Comparison(v2, `GREATER_EQ,
		  (Formula.mk_sum v1 (`Integer Z.one))))))

let rec fold_left_map (f : 'state -> 'i -> 'state*'o)
    (state : 'state) : 'i list -> 'state*'o list =
  function
    | [] -> state, []
    | h::t -> let state', h' = f state h in
      let state'', t' = fold_left_map f state' t in
	state'', (h'::t');;

let rec encode_block (state : translator_state) (block : Cil.block) :
    translator_state * semantics =
	List.fold_left
	  (fun (state1, semantics1) stmt ->
	     let state2, semantics2 = encode_stmt state1 stmt in
	       state2, (compose_semantics semantics1 semantics2))
	   (state, skip_semantics) block.Cil.bstmts

(* TODO: handling of true 'break' instructions still broken *)
and encode_while (state : translator_state) (block : Cil.block) location :
  translator_state * semantics =
  let state' = StringMap.fold (fun var_name _ old_state ->
	fst (get_new_variable_name var_name old_state location))
    state.variable_index state in
  let state'', block_semantics = encode_block state' block in
  let constrained_vars = StringMap.fold (fun before_name cur_var_name prev -> (before_name, (StringMap.find before_name state.variable_index), cur_var_name)::prev)
    state'.variable_index [] in
  let modulo = Formula.formula_true in
  let precondition_vars, preconditions, _ = gen_conditions "loop" (List.map (fun (x, y, z) -> x, z) constrained_vars)
  and _, postconditions, _ = gen_conditions "loop"
    (StringMap.fold (fun name _ prev -> (name, (StringMap.find name state''.variable_index))::prev) state.variable_index []) in
  let initialization_condition =
    Formula.mk_v_and (List.map (fun (_, v1, v2) -> `Atomic(`Comparison((`Variable v1), `EQUAL, (`Variable v2)))) constrained_vars) in
  let stability = Formula.universal_closure_minus
    (Formula.unquantified_to_quantified (Formula.mk_implies (Formula.mk_and (Formula.mk_or initialization_condition preconditions) block_semantics.direct_flow) postconditions))
    (List.append precondition_vars (List.map (fun (_,x,_) -> x) constrained_vars)) in
  let stability_reduced = ((Formula_yices.reduce_dnf_source_formula modulo stability) : Formula.source_dnf_formula :> Formula.source_unquantified_formula) in
    Formula_math.print_source_quantified_formula stdout stability;
    print_newline ();
    Formula_math.print_source_unquantified_formula stdout stability_reduced;
    print_newline ();

    let abstract_transformer =
  StringMap.fold
    (fun vname _ already ->
       let vupbound = vmax "loop" vname in
       let vupbound2 = vupbound ^ "2" in
	 Printf.printf "## invariant for %s: %s\n" vname vupbound;
       let upbound_qformula = `Exists(
	  (List.filter (fun (x : string) -> x <> vupbound) precondition_vars),
	  (Formula.unquantified_to_quantified stability_reduced)) in
       let upbound_formula = (Formula_yices.reduce_dnf_source_formula modulo upbound_qformula  : Formula.source_dnf_formula :> Formula.source_unquantified_formula) in
      let upbound_formula_subst =
	Formula.subst_quantified_formula
	  (fun variable ->
	    if variable = vupbound
	    then `Variable vupbound2
	    else raise Not_found)
	  (Formula.unquantified_to_quantified upbound_formula) in 
      let upbound_formula_minimize =
	`Forall([vupbound2],
	  (Formula.mk_v_and [Formula.unquantified_to_quantified upbound_formula;
	     (Formula.mk_or (`Not upbound_formula_subst)
		(`Atomic(`Comparison((`Variable vupbound),
				     `LESS_EQ,
				     (`Variable vupbound2)))))])) in
      let upbound_formula_optimal =
	(Formula_yices.reduce_dnf_source_formula
	  modulo upbound_formula_minimize 
	   : Formula.source_dnf_formula
	   :> Formula.source_unquantified_formula) in

      Formula_math.print_source_unquantified_formula stdout
	upbound_formula_optimal;
      print_newline();

       let vlowbound = vmin "loop" vname in
       let vlowbound2 = vlowbound ^ "2" in
	 Printf.printf "## invariant for %s: %s\n" vname vlowbound;
       let lowbound_qformula = `Exists(
	  (List.filter (fun (x : string) -> x <> vlowbound) precondition_vars),
	  (Formula.unquantified_to_quantified stability_reduced)) in
       let lowbound_formula = (Formula_yices.reduce_dnf_source_formula modulo lowbound_qformula  : Formula.source_dnf_formula :> Formula.source_unquantified_formula) in
      let lowbound_formula_subst =
	Formula.subst_quantified_formula
	  (fun variable ->
	    if variable = vlowbound
	    then `Variable vlowbound2
	    else raise Not_found)
	  (Formula.unquantified_to_quantified lowbound_formula) in 
      let lowbound_formula_minimize =
	`Forall([vlowbound2],
	  (Formula.mk_v_and [Formula.unquantified_to_quantified lowbound_formula;
	     (Formula.mk_or (`Not lowbound_formula_subst)
		(`Atomic(`Comparison((`Variable vlowbound),
				     `GREATER_EQ,
				     (`Variable vlowbound2)))))])) in
      let lowbound_formula_optimal =
	(Formula_yices.reduce_dnf_source_formula
	  modulo lowbound_formula_minimize 
	   : Formula.source_dnf_formula
	   :> Formula.source_unquantified_formula) in

	print_endline "lowbound qformula";
      Formula_math.print_source_quantified_formula stdout
	lowbound_qformula;
      print_newline();
	print_endline "lowbound formula";
      Formula_math.print_source_unquantified_formula stdout
	lowbound_formula;
      print_newline();
     Formula_math.print_source_unquantified_formula stdout
	lowbound_formula_optimal;
      print_newline();

      Formula.mk_v_and [upbound_formula_optimal;
                        lowbound_formula_optimal; already]
    ) state'.variable_index Formula.formula_true in
    let concrete_approximated_transformer =
      Formula.mk_and preconditions abstract_transformer in
(*
      print_endline "Loop transformer:";
      Formula_math.print_source_unquantified_formula stdout
	concrete_approximated_transformer;
      print_newline();
      print_endline "Exit flow:";
      Formula_math.print_source_unquantified_formula stdout
	block_semantics.break_flow;
      print_newline();
*)
      state', (normal_semantics
		  (Formula.mk_and block_semantics.break_flow concrete_approximated_transformer))
      
and encode_stmt (state : translator_state) (stmt : Cil.stmt) :
    translator_state * semantics =
  match stmt.Cil.skind with
    | Cil.Instr instr_list ->
	List.fold_left
	  (fun (state1, semantics1) instr ->
	     let state2, semantics2 = encode_instr state1 instr in
	       state2, (compose_semantics semantics1 semantics2))
	  (state, skip_semantics) instr_list
    | Cil.Return _ -> state, skip_semantics (* TODO *)
    | Cil.If(bexpr, block1, block2, loc) ->
	let state', bexpr1, bexpr2, formula0 = encode_bexpr state bexpr in
	let state1', formula1 = encode_block state' block1 in
	let state2', formula2 = encode_block state1' block2 in
	let state', f1, f2 = merge_translator_states loc state1' state2' in
	  state',
	(meet_semantics
	   (filter_semantics formula1
	      (Formula.mk_v_and (formula0::bexpr1::f1)))
	   (filter_semantics formula2
	      (Formula.mk_v_and (formula0::bexpr2::f2))))
    | Cil.Loop(block, loc, _, _) ->
	encode_while state block loc
    | Cil.Break _ -> state, break_semantics

and encode_instr (state : translator_state) (instr : Cil.instr) :
  translator_state * semantics =
  match instr with
    | Cil.Set(((Cil.Var dstvar), Cil.NoOffset), expr, location) ->
	let state', expr', formula = encode_expr state expr in
	let state'', v' = get_new_variable_name dstvar.Cil.vname state' location in
	state'',
	(normal_semantics
	(Formula.mk_and formula
	   (`Atomic(`Comparison((`Variable v'), `EQUAL, expr')))))

and encode_bexpr (state : translator_state) (expr : Cil.exp) :
  translator_state * Formula.source_unquantified_formula *
  Formula.source_unquantified_formula * Formula.source_unquantified_formula =
  match expr with
    | Cil.BinOp(Cil.Eq, e1, e2, otype) ->
	encode_comparison state e1 e2 `EQUAL
    | Cil.BinOp(Cil.Ne, e1, e2, otype) ->
	let state', bool_val1, bool_val2,
	  formula = encode_comparison state e1 e2 `EQUAL in
	state', bool_val2, bool_val1, formula
    | Cil.BinOp(Cil.Lt, e1, e2, otype) ->
	encode_comparison state e1 e2 `LESS
    | Cil.BinOp(Cil.Le, e1, e2, otype) ->
	encode_comparison state e1 e2 `LESS_EQ
    | Cil.BinOp(Cil.Gt, e1, e2, otype) ->
	encode_comparison state e1 e2 `GREATER
    | Cil.BinOp(Cil.Ge, e1, e2, otype) ->
	encode_comparison state e1 e2 `GREATER_EQ
    | Cil.UnOp(Cil.LNot, e1, otype) ->
	let state', bool_val1, bool_val2, formula = encode_bexpr state e1 in
	state', bool_val2, bool_val1, formula
    | _ ->
	let state', v_out, formula = encode_expr state expr in
	state',
	(`Atomic(`Comparison(v_out, `EQUAL, (`Integer Z.zero)))),
	(`Not (`Atomic(`Comparison(v_out, `EQUAL, (`Integer Z.zero))))),
	formula
	
and encode_comparison (state : translator_state)
    (e1 : Cil.exp) (e2 : Cil.exp) comparison_op =
  match (Cil.unrollTypeDeep (Cil.typeOf e1)),
    (Cil.unrollTypeDeep (Cil.typeOf e2)) with
    | (Cil.TFloat _), (Cil.TFloat _) ->
      let state1', e1', formula1 = encode_expr state e1 in
      let state2', e2', formula2 = encode_expr state1' e2 in
      let ret = `Atomic(`Comparison(e1', comparison_op, e2')) in
      state2',
      ret,
      (`Not ret),
       (Formula.mk_and formula1 formula2)
    | (Cil.TInt _), (Cil.TInt _) ->
      let state1', e1', formula1 = encode_expr state e1 in
      let state2', e2', formula2 = encode_expr state1' e2 in
      let bool_formula1, bool_formula2 =
        match comparison_op with
	  | `LESS_EQ -> gen_integer_leq e1' e2'
	  | `GREATER_EQ -> gen_integer_leq e2' e1'
	  | `GREATER -> swap (gen_integer_leq e1' e2')
	  | `LESS -> swap (gen_integer_leq e2' e1')
	  | `EQUAL -> gen_integer_equal e1' e2'
	  | `UNEQUAL -> swap (gen_integer_equal e1' e2') in
	state2', bool_formula1, bool_formula2,
	(Formula.mk_and formula1 formula2)
	
    | (Cil.TVoid _), (Cil.TVoid _) -> failwith "comparison between voids???"
    | (Cil.TNamed _), _ | _, (Cil.TNamed _) ->
	failwith "named types after unroll???"
    | _ -> get_top_boolean state

and encode_expr (state : translator_state) (expr : Cil.exp) =
  match expr with
    | Cil.Const(Cil.CInt64(cst, itype, (Some strCst))) ->
	state,
	(`Integer(Z.from_string strCst)),
	Formula.formula_true

    | Cil.Const(Cil.CInt64(cst, itype, None)) ->
	state,
	(`Integer(Z.from_string (Int64.to_string cst))),
	Formula.formula_true

    (* Yields very ugly expressions due to floating-point *)
    | Cil.Const(Cil.CReal(cst, ftype, _)) ->
        let fraction = Q.from_float cst in 
        state,
	(if Z.equal (Q.get_den fraction) Z.one
         then`Integer(Q.get_num fraction)
         else `Quotient(`Integer(Q.get_num fraction),
                        `Integer(Q.get_den fraction))),
	Formula.formula_true
	
    | Cil.BinOp(Cil.PlusA, e1, e2, otype) ->
	let state1', e1', formula1 = encode_expr state e1 in
	let state2', e2', formula2 = encode_expr state e2 in
	state2',
	(Formula.mk_sum e1' e2'),
	(Formula.mk_and formula1 formula2)
	
    | Cil.BinOp(Cil.Mult, e1, e2, otype) ->
	let state1', e1', formula1 = encode_expr state e1 in
	let state2', e2', formula2 = encode_expr state e2 in
	state2',
	(Formula.mk_product e1' e2'),
	(Formula.mk_and formula1 formula2)
	
    | Cil.BinOp(Cil.Div, e1, e2, otype) ->
	let state1', e1', formula1 = encode_expr state e1 in
	let state2', e2', formula2 = encode_expr state e2 in
	state2',
	(`Quotient(e1', e2')),
	(Formula.mk_and formula1 formula2)

    | Cil.BinOp(Cil.MinusA, e1, e2, otype) ->
	let state1', e1', formula1 = encode_expr state e1 in
	let state2', e2', formula2 = encode_expr state e2 in
	state2',
	(Formula.mk_sum e1' (`Neg e2')),
	(Formula.mk_and formula1 formula2)

    | Cil.UnOp(Cil.Neg, e, otype) ->
	let state', e', formula = encode_expr state e in
	state',
	(`Neg e'),
	formula

    | Cil.Lval(Cil.Var(var), Cil.NoOffset) ->
	state,
	(`Variable (get_variable_name state var.Cil.vname)),
	Formula.formula_true

    | Cil.BinOp((Cil.Lt | Cil.Le | Cil.Gt | Cil.Ge | Cil.Eq | Cil.Ne), _, _, _)
    | Cil.UnOp(Cil.LNot, _, _) ->
	let state', bool_val1, bool_val2, formula = encode_bexpr state expr in
	encode_expr_of_bexpr state' bool_val1 bool_val2 formula

    | Cil.CastE(_, e1) -> encode_expr state e1

    | Cil.BinOp(_) -> failwith "unimplemented expr binop"
    | Cil.UnOp(_) -> failwith "unimplemented expr unop"
    | Cil.Const(_) -> failwith "unimplemented expr constant"
    | Cil.Lval(_) -> failwith "unimplemented expr lvalue"
    | Cil.SizeOf(_) -> failwith "unimplemented expr sizeof"
    | Cil.SizeOfE(_) -> failwith "unimplemented expr sizeofE"
    | Cil.SizeOfStr(_) -> failwith "unimplemented expr sizeOfStr"
    | Cil.AlignOf(_) -> failwith "unimplemented expr alignOf"
    | Cil.AlignOfE(_) -> failwith "unimplemented expr alignOfE"
	
    | Cil.AddrOf(_) -> failwith "unimplemented expr addrOf"
    | Cil.StartOf(_) -> failwith "unimplemented expr startOf"

and encode_expr_of_bexpr state bool_val1 bool_val2 formula =
  let state', v_out = get_tmp_variable_name "bool" state in
  state',
  (`Variable v_out),
  (Formula.mk_and formula
     (Formula.mk_or 
	(Formula.mk_and bool_val1
	   (`Atomic(`Comparison((`Variable v_out), `EQUAL, (`Integer Z.one)))))
	(Formula.mk_and bool_val2
	   (`Atomic(`Comparison((`Variable v_out), `EQUAL, (`Integer Z.zero)))))))

let print_abstract_transformer modulo vbound bound_formula_optimal =
      let computed_variable, lc=
	Formula_yices.prepare_projection_linearization_context [ vbound ]
	  (`And[bound_formula_optimal; modulo]) in
      let tree = Piecewise_linear.build_decision_tree lc computed_variable
	(Formula.linearize_unquantified_formula lc modulo)
	(Formula.linearize_unquantified_formula lc bound_formula_optimal) in
      let var_mapper = Formula.var_mapper_of_linearization_context lc in
        print_string vbound;
	print_string " = ";
	Piecewise_linear.print_decision_tree stdout var_mapper "" tree;
	print_newline()
 
let encode_function (fundec : Cil.fundec) =
  Printf.printf "Processing function <%s>\n"
    fundec.Cil.svar.Cil.vname;
  let parameters = collect_variables fundec.Cil.sformals in
  let precondition_vars, preconditions, modulo = gen_preconditions parameters in
  print_string "Preconditions: ";
  Formula_math.print_source_unquantified_formula stdout preconditions;
  print_newline ();
  print_string "Nonemptiness condition: ";
  Formula_math.print_source_unquantified_formula stdout modulo;
  print_newline ();
  Printf.printf "locals:";
  List.iter (function var ->
	       Printf.printf " %s" var)
    (collect_variables fundec.Cil.slocals);
  print_newline ();
  let state', fn_semantics =
    encode_block empty_translator_state fundec.Cil.sbody in
  let formula = fn_semantics.direct_flow in
  let formulas = Formula.mk_and formula preconditions in
  StringMap.iter
    (fun vname vname_end ->
      Printf.printf "## output variable for %s: %s\n" vname vname_end;

      let vupbound = vname ^ "$upbound" in
      let vupbound2 = vupbound ^ "2" in
      let upbound_qformula = Formula.universal_closure_minus
	  (Formula.unquantified_to_quantified
	     (Formula.mk_implies formulas
		(`Atomic(`Comparison((`Variable vname_end),
				     `LESS_EQ,
				     (`Variable vupbound)))))) 
	  (vupbound::precondition_vars) in
      let upbound_formula = ((Formula_yices.reduce_dnf_source_formula
	  modulo upbound_qformula) : Formula.source_dnf_formula :> Formula.source_unquantified_formula) in
      let upbound_formula_subst =
	Formula.subst_quantified_formula
	  (fun variable ->
	    if variable = vupbound
	    then `Variable vupbound2
	    else raise Not_found)
	  (Formula.unquantified_to_quantified upbound_formula) in 
      let upbound_formula_minimize =
	`Forall([vupbound2],
	  (Formula.mk_v_and [Formula.unquantified_to_quantified upbound_formula;
	     (Formula.mk_or (`Not upbound_formula_subst)
		(`Atomic(`Comparison((`Variable vupbound),
				     `LESS_EQ,
				     (`Variable vupbound2)))))])) in
      let upbound_formula_optimal =
	(Formula_yices.reduce_dnf_source_formula
	  modulo upbound_formula_minimize 
	   : Formula.source_dnf_formula
	   :> Formula.source_unquantified_formula) in

      let vlowbound = vname ^ "$lowbound" in
      let vlowbound2 = vlowbound ^ "2" in
      let lowbound_qformula = Formula.universal_closure_minus
	  (Formula.unquantified_to_quantified
	     (Formula.mk_implies formulas
		(`Atomic(`Comparison((`Variable vname_end),
				     `GREATER_EQ,
				     (`Variable vlowbound)))))) 
	  (vlowbound::precondition_vars) in
      let lowbound_formula = ((Formula_yices.reduce_dnf_source_formula
	  modulo lowbound_qformula) : Formula.source_dnf_formula :> Formula.source_unquantified_formula) in
      let lowbound_formula_subst =
	Formula.subst_quantified_formula
	  (fun variable ->
	    if variable = vlowbound
	    then `Variable vlowbound2
	    else raise Not_found)
	  (Formula.unquantified_to_quantified lowbound_formula) in 
      let lowbound_formula_minimize =
	`Forall([vlowbound2],
	  (Formula.mk_v_and [Formula.unquantified_to_quantified lowbound_formula;
	     (Formula.mk_or (`Not lowbound_formula_subst)
		(`Atomic(`Comparison((`Variable vlowbound),
				     `GREATER_EQ,
				     (`Variable vlowbound2)))))])) in
      let lowbound_formula_optimal =
	(Formula_yices.reduce_dnf_source_formula
	  modulo lowbound_formula_minimize 
	   : Formula.source_dnf_formula
	   :> Formula.source_unquantified_formula) in

      Formula_math.print_source_unquantified_formula stdout upbound_formula_optimal;
      Formula_math.print_source_unquantified_formula stdout lowbound_formula_optimal;
      print_newline();
      print_newline();

      Printf.printf "Abstract transformer for function <%s> and variable <%s>:\n"
	fundec.Cil.svar.Cil.vname vname;
      print_abstract_transformer modulo vupbound upbound_formula_optimal;
      print_newline();
      print_abstract_transformer modulo vlowbound lowbound_formula_optimal;
      print_newline ()
    ) state'.variable_index;
  print_newline();;

Projection_apron.init();;

let file = Frontc.parse Sys.argv.(1) ();;

match find_function_in_file file Sys.argv.(2) with
  | None -> print_endline "miaou!!!"
  | Some f -> encode_function f;;
